import Prelude hiding (read, catch)

import HardcodedGameStuff
import PlatformLowlevel
import LogParser
import KolproxyServer
import Logging
import Lua
import State

import KoL.Network
import KoL.Http
import KoL.Connection
import KoL.Command
import KoL.Util
import KoL.UtilTypes
import KoL.Api

import Control.Applicative
import Control.Concurrent
import Control.Exception
import Control.Monad

import Data.Time
import Data.Maybe
import Data.IORef
import qualified Data.ByteString.Char8
import qualified Data.Map

import Network.URI

import Text.Printf (printf)
import Text.JSON

import System.Directory (doesFileExist)
import System.IO

doPROXY_DEBUG _ = return ()
-- doPROXY_DEBUG x = putStrLn $ "PROXY DEBUG: " ++ x

doRetrievePageRawNoScripts = internalKolRequest_pipelining

storeSettingsOnServer ref = do
	doPROXY_DEBUG "storeSettingsOnServer"
	json <- Data.ByteString.Char8.unpack <$> getPageRawNoScripts ("/actionbar.php?action=fetch&for=kolproxy+" ++ kolproxy_version_number ++ "+by+Eleron&format=json") ref
	let Ok jsobj = decodeStrict json :: Result (JSObject JSValue)
	let list = fromJSObject jsobj
	ai <- getApiInfo ref
	stateobj <- makeStateJSON ref (turnsplayed ai)
	let newlist = filter (\(x,_) -> x /= "kolproxy state") list ++ [("kolproxy state", JSObject stateobj)]
	let newjson = encodeStrict $ toJSObject newlist
	postPage "/actionbar.php" [("action", "set"), ("for", "kolproxy " ++ kolproxy_version_number ++ " by Eleron"), ("format", "json"), ("bar", newjson)] ref
	return newjson

loadSettingsFromServer ref = do
	doPROXY_DEBUG "loadSettingsFromServer"
	json <- Data.ByteString.Char8.unpack <$> getPageRawNoScripts ("/actionbar.php?action=fetch&for=kolproxy+" ++ kolproxy_version_number ++ "+by+Eleron&format=json") ref
	let Ok jsobj = decodeStrict json :: Result (JSObject JSValue)
	let list = fromJSObject jsobj
	case lookup "kolproxy state" list of
		Just kolproxystate -> do
			doPROXY_DEBUG $ "loadSettingsFromServer:got " ++ show kolproxystate
			ai <- getApiInfo ref
			doPROXY_DEBUG $ "loadSettingsFromServer:loadStateJSON"
			what <- loadStateJSON ref kolproxystate (turnsplayed ai)
			doPROXY_DEBUG $ "loadSettingsFromServer:mirrorStateIntoDatabase"
			mirrorStateIntoDatabase ref
			doPROXY_DEBUG $ "loadSettingsFromServer:done"
			return $ Just $ show $ what
		Nothing -> return $ Nothing

doProcessPage ref uri params = do
	log_debug_msg ref $ "process page: " ++ (show uri) ++ " [" ++ (show params) ++ "]"

	status_before_func <- getstatusfunc ref
	log_time <- getZonedTime -- TODO: ask CDM for rightnow in API

	let get_the_state ref = do
		cr <- canReadState ref
		case cr of
			False -> return Nothing
			True -> do
				Just (_, y) <- readIORef (state ref)
				return $ Just y

	state_before <- get_the_state ref

	xf <- (rawRetrievePage ref) ref uri params

	mv <- newEmptyMVar

	forkIO $ do
		x <- try $ do
			-- TODO: make sure getstatusfunc is called *right* after the page is completed. Change xf to return it.
			(pagetext, effuri, _hdrs) <- log_time_interval ref ("fetchpage: " ++ (show uri)) $ xf

			status_after_func <- getstatusfunc ref

			let allparams = concat $ catMaybes $ [decodeUrlParams uri, decodeUrlParams effuri, params]
			y <- log_time_interval ref ("processing: " ++ (show uri)) $ runProcessScript ref uri effuri pagetext allparams

			state_after <- get_the_state ref

			forkIO $ (do
				status_before <- status_before_func
				status_after <- status_after_func
				log_page_result ref status_before log_time state_before uri params effuri (Data.ByteString.Char8.unpack pagetext) status_after state_after 
				return ()) `catch` (\e -> putStrLn $ "processpage logging error: " ++ (show (e :: ConnectionException)))
			return (y, pagetext, effuri)
		val <- case x of
			Right (Right msg, _, effuri) -> do
				log_debug_msg ref $ "processed: " ++ (show uri) ++ " [" ++ (show params) ++ "]"
				return $ Right (msg, effuri)
			Right (Left (msg, trace), pagetext, effuri) -> do
				putStrLn $ "Error processing page[" ++ show uri ++ "]: " ++ msg ++ "\n" ++ trace
				return $ Left (add_error_message_to_page ("process-page.lua error: " ++ msg ++ "\n" ++ trace) pagetext, effuri)
			Left e -> do
				putStrLn $ "Exception while processing page[" ++ show uri ++ "]: " ++ (show (e :: SomeException))
				return $ Left (add_error_message_to_page ("process-page.lua exception: " ++ (show e)) (Data.ByteString.Char8.pack "{ Kolproxy page processing. }"), mkuri "/error")
		putMVar mv val

	return $ readMVar mv

doProcessPageWhenever ref uri params = do
	xf <- (rawRetrievePage ref) ref uri params
	return $ do
		(text, effuri, _hdrs) <- xf
		return $ Right (text, effuri)

statusfunc ref = do
	mv <- readIORef $ jsonStatusPageMVarRef $ sessionData $ ref
	return $ do
		x <- modifyMVar mv (\v -> do
			case v of
				Just j -> return (v, j)
				Nothing -> do
					x <- KoL.Api.getStatus ref
					return (Just x, x))
		return x

-- Bugged submitnewchat.php thing
__chat_replace_hack "" = ""
__chat_replace_hack str@(x:xs) = if ((take 6 str) == "</fon>")
	then "</font>" ++ (__chat_replace_hack $ drop 6 str)
	else x:(__chat_replace_hack xs)

kolProxyHandlerWhenever uri params baseref = do
	let ref = baseref {
		processingstuff_ = ProcessingRefStuff {
			processPage_ = doProcessPageWhenever,
			rawRetrievePage_ = (\ref uri params -> doRetrievePageRawNoScripts ref uri params False),
			nochangeRawRetrievePageFunc_ = doRetrievePageRawNoScripts,
			getstatusfunc_ = statusfunc
		}
	}
	Right (text, effuri) <- join $ (processPage ref) ref uri params
	resptext <- case uriPath uri of
		"/newchatmessages.php" -> do
			let allparams = concat $ catMaybes $ [decodeUrlParams uri, decodeUrlParams effuri, params]
			x <- runChatScript ref uri effuri text allparams
			case x of
				Right msg -> return msg
				Left (msg, trace) -> do
					putStrLn $ "chat error: " ++ (msg ++ "\n" ++ trace)
					return text
		"/submitnewchat.php" -> return $ Data.ByteString.Char8.pack $ __chat_replace_hack $ Data.ByteString.Char8.unpack $ text
		_ -> return text
	resp <- makeResponseWithNoExtraHeaders resptext effuri []
	return resp

handleRequest ref uri effuri headers params pagetext = do
	log_debug_msg ref $ printf "handlereq: got uri: %s | %s (from %s), size %d, params: %s" (uriPath effuri) (uriQuery effuri) (show uri) (Data.ByteString.Char8.length pagetext) (show params)

	let allparams = concat $ catMaybes $ [decodeUrlParams uri, decodeUrlParams effuri, params]

	xresptext <- log_time_interval ref ("printing: " ++ (show uri)) $ runPrinterScript ref uri effuri pagetext allparams

	let resptext = case xresptext of
		Right msg -> msg
		Left (msg, trace) -> add_error_message_to_page ("printer.lua error: " ++ msg ++ "\n" ++ trace) pagetext

	log_time_interval ref ("making response") $ makeResponse resptext effuri headers


make_ref baseref = do
	let doRetrievePageRaw ref url params = do
		log_file_retrieval ref url params
		xf <- (nochangeRawRetrievePageFunc ref) ref url params True
		mv <- newEmptyMVar
		forkIO $ do
			x <- try $ xf
			putMVar mv x
		return $ do
			x <- readMVar mv
			case x of
				Right r -> return r
				Left e -> throwIO (e :: SomeException)
	let ref = baseref {
		processingstuff_ = ProcessingRefStuff {
			processPage_ = doProcessPage,
			rawRetrievePage_ = doRetrievePageRaw,
			nochangeRawRetrievePageFunc_ = doRetrievePageRawNoScripts,
			getstatusfunc_ = statusfunc
		}
	}
	state_is_ok <- (do
		refreshstatus ref
		loadState ref
		return True) `catch` (\e -> do
			putStrLn $ "loadstate exception: " ++ show (e :: SomeException)
			return False)
	return ref { stateValid_ = state_is_ok }

kolProxyHandler uri params baseref = do
	let _fake_log_ref = baseref
	log_debug_msg _fake_log_ref $ "proxy handler: " ++ (show uri) ++ " [" ++ (show params) ++ "]" -- {" ++ (show cookiebundle) ++ "}"
	forkIO $ update_data_files -- TODO: maybe not for *every single page*?

	ref <- log_time_interval _fake_log_ref ("make ref for: " ++ (show uri)) $ make_ref baseref

	let allparams = concat $ catMaybes $ [decodeUrlParams uri, params]

	let check_pwd ref = do
		ai <- getApiInfo ref
		let realpwd = (pwd ai)
		return (lookup "pwd" allparams == Just realpwd)

	do
		canread <- canReadState ref
		log_debug_msg ref $ "proxy handler: state{" ++ (show canread) ++ "}, allparams [" ++ (show allparams) ++ "]"

	let getpage = do
		log_debug_msg ref $ "return page: " ++ (show uri) ++ " [" ++ (show params) ++ "]"
		join $ (processPage ref) ref uri params

	let response = case uriPath uri of
		"/login.php" -> case params of
			Nothing -> Nothing
			Just p_sensitive -> Just $ do
				(pt, effuri, hdrs) <- case lookup "password" p_sensitive of
					Just "" -> do
						putStrLn $ "Logging in..."
						postLoginPage uri p_sensitive (ref, (useragent $ connection $ ref), (hostUri $ connection $ ref))
					_ -> do
						putStrLn $ "Logging in over https..."
						postHttpsLoginPage uri p_sensitive (ref, (useragent $ connection $ ref), (hostUri $ connection $ ref))
				p_sensitive <- return ()
				putStrLn $ "Login fetch " ++ (show $ uriPath uri) ++ ", gave " ++ (show $ uriPath effuri)
				log_debug_msg ref $ "kolproxy:login.php (handling request, logging in)"
				let new_cookie = lookup "Set-Cookie" hdrs
				putStrLn $ "new_cookie: " ++ (show new_cookie)
				if (new_cookie == Nothing)
					then do
						putStrLn $ "Error: No cookie from logging in!"
						putStrLn $ "  headers: " ++ (show hdrs)
						putStrLn $ "  url: " ++ (show effuri)
-- 						putStrLn $ "pt: " ++ pt
						handleRequest ref uri effuri hdrs params pt
					else (do
-- 						ref <- return $ ref { otherstuff_ = (otherstuff_ ref) { connection_ = (connection_ $ otherstuff_ $ ref) { cookie = new_cookie } } }
						ref <- make_ref $ baseref { otherstuff_ = (otherstuff_ ref) { connection_ = (connection_ $ otherstuff_ $ ref) { cookie = new_cookie } } }
						putStrLn $ "login.php getting state"
						ai <- getApiInfo ref
						 -- "rollover" is not set yet. It's null, due to broken game API
						putStrLn $ "New login: " ++ show (get_sessid new_cookie, hostUri $ connection $ ref, charName ai, ascension ai)
-- 						writeIORef (state ref) (Just new_statevalue)
-- 						saveState ref
						putStrLn $ "Logging in as " ++ (charName ai) ++ " (ascension " ++ (show $ ascension ai) ++ ")"
						what <- loadSettingsFromServer ref
						putStrLn $ "loaded settings: " ++ (fromMaybe "nothing" what)

						forkIO $ compressLogs (charName ai) (ascension ai)

						putStrLn $ "login.php redirect"
						putStrLn $ "  hdrs: " ++ show hdrs
						putStrLn $ "  pt: " ++ (Data.ByteString.Char8.unpack pt)
						makeRedirectResponse pt uri (hdrs ++ [("Location", "game.php")])) `catch` (\e -> do
							putStrLn $ "Error: Failed to read state. This can happen if you try to log in for the first time in a day while in combat, in a choice adventure or in valhalla."
							putStrLn $ "Login error: " ++ (show (e :: Control.Exception.SomeException))
							-- TODO: Eeek! When this happens, there's no proper session, stuff gets fucked
							makeRedirectResponse pt uri hdrs)
-- 							return $ Response (5,0,0) "" (hdrs ++ [Header HdrContentType "text/html; charset=UTF-8", Header HdrCacheControl "no-cache"]) "Error: Failed to log in. This can happen if you try to log in for the first time in a day while in combat, in a choice adventure or in valhalla.")
--					_ -> makeResponse (Data.ByteString.Char8.pack "Non-secure login rejected. Non-secure login sends your password in cleartext and is a really bad idea, especially for proxy connections. Enable javascript to use KoL's safer hash-based login system!") (mkuri "/login-denied") []


-- TODO: intercept if logged in before page. automate if logged in after page. support logging both.

		"/custom-clear-lua-script-cache" -> Just $ do
			okpwd <- check_pwd ref
			if okpwd
				then do
					writeIORef (luaInstances $ sessionData $ ref) Data.Map.empty
					makeResponse (Data.ByteString.Char8.pack $ "Cleared Lua script cache.") uri []
				else makeResponse (Data.ByteString.Char8.pack $ "Invalid pwd field") uri []

		"/custom-store-settings" -> Just $ do
			okpwd <- check_pwd ref
			if okpwd
				then do
					newjson <- storeSettingsOnServer ref
					makeResponse (Data.ByteString.Char8.pack $ "Stored " ++ (show $ length newjson) ++ " bytes.") uri []
				else makeResponse (Data.ByteString.Char8.pack $ "Invalid pwd field") uri []

		"/custom-load-settings" -> Just $ do
			okpwd <- check_pwd ref
			if okpwd
				then do
					what <- loadSettingsFromServer ref
					case what of
						Just msg -> makeResponse (Data.ByteString.Char8.pack $ "Settings loaded: " ++ msg ++ ".") uri []
						Nothing -> makeResponse (Data.ByteString.Char8.pack $ "Error: No kolproxy settings stored on server.") uri []
				else makeResponse (Data.ByteString.Char8.pack $ "Invalid pwd field") uri []

		"/custom-logs" -> Just $ do
			let which = matchGroups "which=([^.]+)" (uriQuery uri)
			pt <- case which of
				[[filename]] -> showLogs (Just filename)
				_ -> showLogs Nothing
			makeResponse (Data.ByteString.Char8.pack pt) uri []

		"/custom-settings" -> Just $ do
-- 			putStrLn $ "custom-settings # " ++ show allparams
			okpwd <- check_pwd ref
			if okpwd
				then do
					case lookup "action" allparams of
						Nothing -> return ()
						Just "set state" -> do
							case (lookup "stateset" allparams, lookup "name" allparams, lookup "value" allparams) of
								(Just stateset, Just name, Just value) -> do
-- 									putStrLn $ "  setting " ++ stateset ++ ":" ++ name ++ " = " ++ value
									if value == ""
										then unsetState ref stateset name
										else setState ref stateset name value
-- 									saveState ref
								_ -> return ()
						Just x -> throwIO $ InternalError $ "Custom settings action not recognized: " ++ x
					handleRequest ref uri uri [] params (Data.ByteString.Char8.pack "Empty page.")
				else makeResponse (Data.ByteString.Char8.pack "Invalid pwd field") uri []

		"/kolproxy-automation-script" -> Just $ do
			okpwd <- check_pwd ref
			if okpwd
				then do
					runauto <- getState ref "character" "run automation scripts"
					if (runauto == Just "no")
						then makeResponse (Data.ByteString.Char8.pack "Automation is currently disabled (scripts are not allowed to download pages). Go to the kolproxy settings to change it.") uri []
						else do
							(pt, effuri, hdrs) <- do
								let reqtype = if isJust params then "POST" else "GET"
								log_debug_msg ref $ "run-intercept: " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"
								zz <- runInterceptScript ref uri allparams reqtype
								log_debug_msg ref $ "ran-intercept: " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"
								case zz of
									Right (text, effuri) -> return (text, effuri, [])
									Left (msg, trace) -> do
										putStrLn $ "intercept.lua error: " ++ msg ++ "\n" ++ trace
										return (add_error_message_to_page ("intercept.lua error: " ++ msg ++ "\n" ++ trace) (Data.ByteString.Char8.pack "{ Kolproxy automation script. }"), mkuri "/error", [])
							log_debug_msg ref $ "handling: " ++ (show effuri) ++ "  <- " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"
							handleRequest ref uri effuri hdrs params pt
				else makeResponse (Data.ByteString.Char8.pack "Invalid pwd field") uri []
		_ -> Nothing

	retresp <- log_time_interval ref ("run handler for: " ++ (show uri)) $ case response of
		Just r -> do
			canread <- canReadState ref
			if and [not canread, uriPath uri /= "/login.php", uriPath uri /= "/afterlife.php"]
				then do
					putStrLn $ "Error: Can't read state! Don't log in for the first time in a day while in a fight or choice noncombat, and don't log in while in valhalla!"
					gp <- getpage
					case gp of
						Left (pt, effuri) -> makeResponse pt effuri []
						Right (pt, effuri) -> do
							if (uriPath effuri) `elem` ["/fight.php", "/choice.php"]
								then makeResponse (add_error_message_to_page "Error: kolproxy can't read state!" pt) effuri []
								else handleRequest ref uri effuri [] params pt
				else r
		Nothing -> do
			log_debug_msg ref $ "page not specifically handled: " ++ (show uri)
			canread_before <- canReadState ref
-- 			putStrLn $ "canread_before: " ++ show (uri, canread_before)

			-- TODO: Move to specific handler?
			when (uriPath uri == "/logout.php" && canread_before) $ do
				newjson <- storeSettingsOnServer ref
				putStrLn $ "stored settings on server (" ++ (show $ length newjson) ++ " bytes.)"

			should_run_intercept_script <- if canread_before
				then do
					x <- getState ref "character" "run automation scripts"
					return $ x /= Just "no"
				else return False

			downloaded_page <- if should_run_intercept_script
				then do
					let reqtype = if isJust params then "POST" else "GET"
					log_debug_msg ref $ "run-intercept: " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"
					zz <- log_time_interval ref ("intercepting: " ++ (show uri)) $ runInterceptScript ref uri allparams reqtype
					log_debug_msg ref $ "ran-intercept: " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"
					case zz of
						Right (pt, effuri) -> return $ Right (pt, effuri)
						Left (msg, trace) -> do
							putStrLn $ "intercept.lua error: " ++ msg ++ "\n" ++ trace
							return $ Left (add_error_message_to_page ("intercept.lua error: " ++ msg ++ "\n" ++ trace) (Data.ByteString.Char8.pack "{ No page loaded. }"), mkuri "/error")
				else log_time_interval ref ("run getpage for: " ++ (show uri)) $ getpage

			(new_page, ref) <- case downloaded_page of
				Left dl -> return (Left dl, ref)
				Right (pt, effuri) -> do
					ref <- if canread_before && (uriPath effuri /= "/afterlife.php")
						then return ref
						else log_time_interval _fake_log_ref ("make ref-2 for: " ++ (show uri)) $ make_ref baseref
					canread_after <- canReadState ref
-- 					putStrLn $ "canread_after: " ++ show (uri, canread_after)
					should_run_automate_script <- if canread_after
						then do
							x <- getState ref "character" "run automation scripts"
							return $ x /= Just "no"
						else return False
					x <- if should_run_automate_script
						then do
							log_debug_msg ref $ "run-automate: " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"
							y <- log_time_interval ref ("automating: " ++ (show uri)) $ runAutomateScript ref uri effuri pt allparams
							log_debug_msg ref $ "ran-automate: " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"
							case y of
								Right autotext -> return $ Right (autotext, effuri)
								Left (msg, trace) -> do
									putStrLn $ "automate.lua error: " ++ msg ++ "\n" ++ trace
									return $ Left (add_error_message_to_page ("automate.lua error: " ++ msg ++ "\n" ++ trace) pt, effuri)
						else return $ Right (pt, effuri)
					return (x, ref)

			do
				let effuri = case new_page of
					Left (_, x) -> x
					Right (_, x) -> x
				log_debug_msg ref $ "handling: " ++ (show effuri) ++ "  <- " ++ (show uri) ++ " [" ++ (show allparams) ++ "]"

			case new_page of
				Left (pt, effuri) -> makeResponse pt effuri []
				Right (pt, effuri) -> log_time_interval ref ("run handle request for: " ++ (show uri)) $ handleRequest ref uri effuri [] params pt
	return retresp

main = platform_init $ do
	hSetBuffering stdout LineBuffering
	have_process_page <- doesFileExist "scripts/process-page.lua"
	if have_process_page
		then putStrLn $ "Starting..."
		else do
			putStrLn $ "WARNING: Trying to start without required files in the \"scripts\" directory."
			putStrLn $ "         Did you unzip the files correctly?"
	portenv <- getEnvironmentSetting "KOLPROXY_PORT"
	let portnum = case portenv of
		Just x -> fromJust $ read_as x :: Integer
		Nothing -> 18481
	runProxyServer kolProxyHandler kolProxyHandlerWhenever portnum `catch` (\e -> putStrLn ("mainError: " ++ show (e :: Control.Exception.SomeException)))
	putStrLn $ "Done! (main finished)"
